(** Template file and exercises for Pactole *)

(* Helping typeclass resolution avoid infinite loops. *)
Typeclasses eauto := (bfs).

(* The general Pactole framework. *)
Require Import Pactole.Setting.

(* The space in which robots evolve.  (in subdirectory Spaces/)
   Possible choices include: R, R2, Ring, ... *)
Require Import Pactole.Spaces.R2.
Instance Loc : Location := make_Location R2.

(* The type of observations.  (in subdirectory Observations/)
   Possible choices include: multiset, set, their variants with limited radius, ... *)
Require Import Pactole.Observations.MultisetObservation.

(* Optionally, you can add constraint on the model such as rigid move, or using simlarities. *)
Require Import Pactole.Models.Similarity.
Require Import Pactole.Models.Rigid.
Close Scope R_scope.


Section Template.

(* Define the number of robots: first the number of good ones, then the number of byzantine ones. *)
Variable n : nat.
Hypothesis n_non_0 : n <> 0%nat.
Instance MyRobots : Names := Robots (2 * n) 0.
(* Here, we have [2 * n] good robots and no byzantine ones, with [n] arbitrary but non null. *)

Instance RC : robot_choice location := { robot_choice_Setoid := location_Setoid }.

(* Define what the space and the state of a robot are.
   Here the state only contains the location of the robot, which is in the real plane.
   Furthermore, we  *)
Instance Info : State location := OnlyLocation
  (fun f => sigT (fun sim : Similarity.similarity location => Bijection.section sim == f)).

(* What can demons choose to change of frame of reference?
   Here we use similarities. *)
Instance FC : frame_choice (Similarity.similarity location) := FrameChoiceSimilarity.

(* What choices can demons make for the state update (for both active and inactive robots)?
   Here they do not make any choice, hence we use [unit]. *)
Instance UC : update_choice unit := {update_choice_EqDec := unit_eqdec}.
Instance IC : inactive_choice unit := {inactive_choice_EqDec := unit_eqdec}.


(* Finally, define what the update functions are. *)
Instance UpdateFun : update_function location (Similarity.similarity location) unit := {
  update := fun _ _ _ pt _ => pt }.
Proof. now repeat intro. Defined.

Instance InactiveFun : inactive_function unit := {
  inactive := fun config id _ => config id }.
Proof. now repeat intro; subst. Defined.

(* If you have constraints on your model, now is the time to prove them.
   Here, we prove that our update function is indeed rigid. *)
Check rigid_update.
Instance Update : RigidSetting.
Proof. split. now intros. Qed.


(* The setup is now complete, we can start working.
   Here we have a dummy example where robot never move, and thus they solve the stop problem. *)

Definition stall (e : execution) := Stream.hd e == Stream.hd (Stream.tl e).

Definition my_problem : execution -> Prop := fun e => Stream.forever stall e.

(* This is the robogram itself: move to the origin, which is the current location of the observing
    robot so that it amounts to not moving. *)
Definition my_robogram (s : observation) : location := origin.

(* We must prove that equal spectra gives equal destinations.
   This is non trivial as both equalities are arbitrary equivalence relations. *)
Lemma my_robogram_compat : Proper (equiv ==> equiv) my_robogram.
Proof. repeat intro. reflexivity. Qed.

(* This combines the robogram and its compatibility proof. *)
Definition full_robogram : robogram := {|
  pgm := my_robogram;
  pgm_compat := my_robogram_compat |}.

(* Here is the final result that we could try to prove. *)
Theorem my_solution : forall d config,
  Fair d ->
  my_problem (execute full_robogram d config).
Proof.
(* This statement is only here to illustrate the framework.
   We actually lack an hypothesis to prove it.
   Actual exercises follow. *)
Abort.


(** *     Exercises     **)

(** ***  Exercise 1: Modelling  **)

(** Define the following problems as properties over executions.
    You can define auxiliary definitions first. *)
Check Stream.eventually.
Check Stream.forever.
Check Stream.instant.
Print ident.

(** Convergence: given any [eps > 0], all good robots end up within a disk of radius eps. *)
Definition convergence (e : execution) := .

(** Gathering: all robots end up in exactly the same location. *)
Definition gathering (e : execution) := .

(** Exploration with stop: all locations are visited then all robots stop (no longer move).

   It does not really makes sense in a continuous space such as the current one
   but the definition would be the same in a ring or graph. *)
Definition Explore_and_Stop e := .

(** Perpetual exploration: each location is visited infinitely often. *)
Definition perpetual_exploration e := .

(** Safety: stay outside of a given set [danger] of states. *)
Definition safe danger e := .

(** ***  Exercise 2: Properties of streams and demons  **)

(** Define the until and weak until operators on streams.

    [P until Q] is valid if P holds up to a point where [Q] holds (which must exists).

    [P weak_until Q] is valid if either [P until Q] holds or [P] holds forever. *)
Print Stream.t.
Inductive until (P Q : execution -> Prop) : execution -> Prop := .

Definition weak_until (P Q : execution -> Prop) : execution -> Prop := .

(** Define properties on demons:
   1) a fully-synchronous demon (all robots move in each round)
   2) a centralized demon (at most a single robot moves in each round) *)
(* The top 5 fields remind you of the decision taken by the demon, the rest is properties. *)
Print demonic_action.

Definition fullySynchronous (d : demon) := .

Definition centralized (d : demon) := .
End Template.

(** ***  Exercise 3: Lights  **)

(** a) Copy the template above (mostly the state of robots) and modify it to introduce lights,
       that is, finite information visible from other robots.

    b) Update your solution to modify any State instance by adding lights to it. *)
Section WithLights.
Context `{Location}.
Context {nbLights : nat}.
Definition lights := fin nbLights.

Instance lights_Setoid : Setoid lights := sig_Setoid (eq_setoid nat).
Instance lights_EqDec : EqDec lights_Setoid := sig_EqDec nat_eq_eqdec _.

Instance WithLights {T} (St : State T) : State (T * lights) := {|
  get_location := fun x => get_location (fst x);
  state_Setoid := prod_Setoid (@state_Setoid _ _ St) lights_Setoid;
  precondition := fun f => precondition f;
  lift := fun f x => (lift f (fst x), snd x) |}.
Proof.
+ abstract (intros Hid [x l]; simpl; split; trivial; []; apply (lift_id Hid)).
+ abstract (intros f x; simpl fst; apply get_location_lift).
+ abstract (intros [] [] []; simpl in *; now apply get_location_compat).
+ abstract (intros f g Hfg [] [] []; simpl in *; split; trivial; []; now apply lift_compat).
Defined.
End WithLights.

(** ***  Exercise 4: handling a full problem  **)

(** Formalize and prove an algorithm to solve rendezvous in R² with 1 bit
    from "Rendezvous of two robots with visible bits", G. Viglietta, ALGOSENSORS 2013

    2 full lights: A and B

    Robogram:
    - me=A, you=A: move to middle location, light goes to B
    - me=A, you=B: move to the other location
    - me=B, you=A: stay
    - me=B, you=B: stay, light goes to A

    Under which synchrony hypothesis (FSYNC, SSYNC, ASYNC, fairness) is it valid? *)
Require Import Pactole.Models.Similarity.
Require Import Lia.

Section RDV.

(* Avoid problems with the definitions above. *)
Remove Hints RC Info FC UC IC eq_setoid multiset_observation : typeclass_instances.

Definition rendezvous : robogram := .

Theorem Viglietta_correct : forall d config, gathering (execute rendezvous d config).
Proof.
Qed.

End RDV.
